#include <WiFi.h>
#include <WiFiMulti.h>
#include <OneButton.h>
#include <HTTPClient.h>
#include <Wire.h>
#include "freertos/event_groups.h"
#include "esp_camera.h"
#include "esp_wifi.h"
#include "SSD1306.h"
#include "OLEDDisplayUi.h"
#include "ca.h"

/***************************************
 *  WiFi
 **************************************/
#define WIFI_SSID   ""
#define WIFI_PASSWD ""

/***************************************
 *  Slack
 **************************************/
#define SLACK_TOKEN ""
#define SLACK_CHANNEL ""

/***************************************
 *  PinOUT
 **************************************/
#define PWDN_GPIO_NUM       26
#define RESET_GPIO_NUM      -1
#define XCLK_GPIO_NUM       32
#define SIOD_GPIO_NUM       13
#define SIOC_GPIO_NUM       12

#define Y9_GPIO_NUM         39
#define Y8_GPIO_NUM         36
#define Y7_GPIO_NUM         23
#define Y6_GPIO_NUM         18
#define Y5_GPIO_NUM         15
#define Y4_GPIO_NUM         4
#define Y3_GPIO_NUM         14
#define Y2_GPIO_NUM         5
#define VSYNC_GPIO_NUM      27
#define HREF_GPIO_NUM       25
#define PCLK_GPIO_NUM       19
#define AS312_PIN           33
#define BUTTON_1            34

#define I2C_SDA             21
#define I2C_SCL             22

#define SSD130_MODLE_TYPE   GEOMETRY_128_64

/***************************************
 *  Display
 **************************************/
#define SSD1306_ADDRESS 0x3c
SSD1306 oled(SSD1306_ADDRESS, I2C_SDA, I2C_SCL, SSD130_MODLE_TYPE);
OLEDDisplayUi ui(&oled);

String ip;
EventGroupHandle_t evGroup;
OneButton button1(BUTTON_1, true);

static esp_err_t capture_handler();

static void drawFrame1(OLEDDisplay *display, OLEDDisplayUiState *state, int16_t x, int16_t y);
static void drawFrame2(OLEDDisplay *display, OLEDDisplayUiState *state, int16_t x, int16_t y);

FrameCallback frames[] = {drawFrame1, drawFrame2};
#define FRAMES_SIZE (sizeof(frames) / sizeof(frames[0]))

#define JST 3600*9 // 時刻設定用

WiFiMulti WiFiMulti; // WiFi


static void buttonClick()
{
    capture_handler();
}

static void buttonLongPress()
{
    int x = oled.getWidth() / 2;
    int y = oled.getHeight() / 2;
    ui.disableAutoTransition();
    oled.setTextAlignment(TEXT_ALIGN_CENTER);
    oled.setFont(ArialMT_Plain_10);
    oled.clear();

    if (PWDN_GPIO_NUM > 0) {
        pinMode(PWDN_GPIO_NUM, PULLUP);
        digitalWrite(PWDN_GPIO_NUM, HIGH);
    }
    oled.drawString(x, y, "Press again to wake up");
    oled.display();

    delay(6000);

    oled.clear();
    oled.display();
    oled.displayOff();

    esp_sleep_enable_ext0_wakeup((gpio_num_t )BUTTON_1, LOW);
    esp_deep_sleep_start();
}


static void drawFrame1(OLEDDisplay *display, OLEDDisplayUiState *state, int16_t x, int16_t y)
{
    display->setTextAlignment(TEXT_ALIGN_CENTER);

    int y1 = 25;

    if (oled.getHeight() == 32) {
        y1 = 10;
    }

    display->setFont(ArialMT_Plain_16);
    display->drawString(64 + x, y1 + y, ip);

    if (digitalRead(AS312_PIN)) {
        display->drawString(64 + x, 5 + y, "AS312 Trigger");

        static time_t old_time = 0;
        time_t new_time = time(NULL);

        // 検知したあと5秒間は待つ
        if (old_time + 5 < new_time) {
            Serial.println("Detection AS312...");
            capture_handler();
        }
        old_time = new_time;
    }
}

static void drawFrame2(OLEDDisplay *display, OLEDDisplayUiState *state, int16_t x, int16_t y)
{
    display->setTextAlignment(TEXT_ALIGN_CENTER);
    display->setFont(ArialMT_Plain_10);

    if (oled.getHeight() == 32) {
        display->drawString( 64 + x, 0 + y, "Camera Ready! Use");
        display->drawString(64 + x, 10 + y, "http://" + ip );
        display->drawString(64 + x, 16 + y, "to connect");
    } else {
        display->drawString( 64 + x, 5 + y, "Camera Ready! Use");
        display->drawString(64 + x, 25 + y, "http://" + ip );
        display->drawString(64 + x, 45 + y, "to connect");
    }
}

static void setClock() {
    configTime( JST, 0, "ntp.nict.jp", "ntp.jst.mfeed.ad.jp");

    Serial.print(F("Waiting for NTP time sync: "));
    time_t nowSecs = time(nullptr);
    while (nowSecs < 8 * 3600 * 2) {
        delay(500);
        Serial.print(F("."));
        yield();
        nowSecs = time(nullptr);
    }

    Serial.println();
    struct tm timeinfo;
    gmtime_r(&nowSecs, &timeinfo);
    Serial.print(F("Current time: "));
    Serial.print(asctime(&timeinfo));
}



static int slack_post_request (const char *buf, int len)
{
    HTTPClient https;
    int httpCode = 0;
    const char* host = "slack.com";  // host URL

    // url作成
    String url = "/api/files.upload";
    url += "?token=" + (String)SLACK_TOKEN;
    url += "&filename=esp32%2ejpg";
    url += "&filetype=jpg";
    url += "&initial_comment=detected!";
    url += "&esp32%2ejpg";
    url += "&channels=%23" + (String)SLACK_CHANNEL;

    if (https.begin(host, 443, url.c_str(), rootCACertificate)) {

        int content_size = 0;
        int malloc_size = len * 2; // ヘッダの付与などを行うためすこし大きめに取得しておく
        char *content = (char *)ps_malloc(malloc_size);
        if (content == NULL)
        {
            Serial.println("ps_malloc failed!");
            return -1;
        }

        // content作成
        String boundary = "----1cfd0b6e934fe824";
        https.addHeader("Content-Type", "multipart/form-data; boundary=" + boundary );

        content_size += sprintf(content + content_size, "--%s\r\n", boundary.c_str());
        content_size += sprintf(content + content_size, "Content-Disposition: form-data; name=\"file\"; filename=\"esp32cap.jpg\"\r\n");
        content_size += sprintf(content + content_size, "Content-Type: image/jpeg\r\n");
        content_size += sprintf(content + content_size, "\r\n");
        memcpy(content + content_size, buf, len);
        content_size += len;
        content_size += sprintf(content + content_size, "\r\n");
        content_size += sprintf(content + content_size, "--%s--\r\n", boundary.c_str());

        httpCode = https.POST((uint8_t *)content, content_size);
        if (httpCode > 0) {
            Serial.printf("[HTTP] POST... code: %d\n", httpCode);
            if(httpCode == HTTP_CODE_OK) {
                Serial.printf("[HTTP] POST OK... code: %d\n", httpCode);
            }
            else {
                Serial.printf("[HTTP] POST... failed, error: %s\n", https.errorToString(httpCode).c_str());
            }
        }
        https.end();
        free(content);
    }
    else {
        Serial.println("http start Error.. ");
    }

    return httpCode;
}

static esp_err_t capture_handler()
{
    camera_fb_t *fb = NULL;
    esp_err_t res = ESP_OK;
    int64_t fr_start = esp_timer_get_time();

    fb = esp_camera_fb_get();
    if (!fb) {
        Serial.printf("Camera capture failed");
        return ESP_FAIL;
    }

    slack_post_request((const char *)fb->buf, fb->len);

    esp_camera_fb_return(fb);
    int64_t fr_end = esp_timer_get_time();
    Serial.printf("JPG: %uB %ums", (uint32_t)(fb->len), (uint32_t)((fr_end - fr_start) / 1000));
    return res;
}

static void camera_init_config()
{
    // Camera Init
    camera_config_t config;
    config.ledc_channel = LEDC_CHANNEL_0;
    config.ledc_timer = LEDC_TIMER_0;
    config.pin_d0 = Y2_GPIO_NUM;
    config.pin_d1 = Y3_GPIO_NUM;
    config.pin_d2 = Y4_GPIO_NUM;
    config.pin_d3 = Y5_GPIO_NUM;
    config.pin_d4 = Y6_GPIO_NUM;
    config.pin_d5 = Y7_GPIO_NUM;
    config.pin_d6 = Y8_GPIO_NUM;
    config.pin_d7 = Y9_GPIO_NUM;
    config.pin_xclk = XCLK_GPIO_NUM;
    config.pin_pclk = PCLK_GPIO_NUM;
    config.pin_vsync = VSYNC_GPIO_NUM;
    config.pin_href = HREF_GPIO_NUM;
    config.pin_sscb_sda = SIOD_GPIO_NUM;
    config.pin_sscb_scl = SIOC_GPIO_NUM;
    config.pin_pwdn = PWDN_GPIO_NUM;
    config.pin_reset = RESET_GPIO_NUM;
    config.xclk_freq_hz = 20000000;
    config.pixel_format = PIXFORMAT_JPEG;
    //init with high specs to pre-allocate larger buffers
    if (psramFound()) {
        config.frame_size = FRAMESIZE_UXGA;
        config.jpeg_quality = 10;
        config.fb_count = 2;
    } else {
        config.frame_size = FRAMESIZE_SVGA;
        config.jpeg_quality = 12;
        config.fb_count = 1;
    }

    esp_err_t err = esp_camera_init(&config);
    if (err != ESP_OK) {
        Serial.printf("Camera init Fail");
        oled.clear();
        oled.drawString(oled.getWidth() / 2, oled.getHeight() / 2, "Camera init Fail");
        oled.display();
        while (1);
    }
}

/****************************************************
 * setup
 */

void setup()
{
    // Serial Setting
    Serial.begin(115200);
    Serial.setDebugOutput(true);
    Serial.println();

    // Memory Check
    #if defined(BOARD_HAS_PSRAM)
    Serial.printf("Board has PSRAM\n");
    #else
    Serial.printf("Board doesn't have PSRAM\n");
    #endif
    Serial.println("Heap: ");
    Serial.println(ESP.getFreeHeap());
    Serial.println("PSRAM: ");
    Serial.println(ESP.getPsramSize());

    // PIR Setting
    pinMode(AS312_PIN, INPUT);
    if (I2C_SDA > 0)
        Wire.begin(I2C_SDA, I2C_SCL);

    // Display Setting
    int x = oled.getWidth() / 2;
    int y = oled.getHeight() / 2;
    oled.init();
    Wire.setClock(100000);  //! Reduce the speed and prevent the speed from being too high, causing the screen
    oled.setFont(ArialMT_Plain_16);
    oled.setTextAlignment(TEXT_ALIGN_CENTER);
    delay(50);
    oled.drawString(x, y - 10, "TTGO Camera");
    oled.display();

    if (!(evGroup = xEventGroupCreate())) {
        Serial.println("evGroup Fail");
        while (1);
    }
    xEventGroupSetBits(evGroup, 1);

    // Camera setting
    camera_init_config();

    //drop down frame size for higher initial frame rate
    sensor_t *s = esp_camera_sensor_get();
    s->set_framesize(s, FRAMESIZE_QVGA);
    s->set_vflip(s, 1); //TTGOの場合レンズが上下反転しているため垂直方向にフリップさせておく

    // Button Setting
    button1.attachLongPressStart(buttonLongPress);
    button1.attachClick(buttonClick);

    // WiFi Setting
    WiFi.mode(WIFI_STA);
    WiFiMulti.addAP(WIFI_SSID, WIFI_PASSWD);

    while ((WiFiMulti.run() != WL_CONNECTED)) {
        delay(500);
        Serial.print(".");
    }
    Serial.println("");
    Serial.println("WiFi connected");

    // SSL通信のため時刻設定しておく
    setClock();

    delay(50);

    ui.setTargetFPS(30);
    ui.setIndicatorPosition(BOTTOM);
    ui.setIndicatorDirection(LEFT_RIGHT);
    ui.setFrameAnimation(SLIDE_LEFT);
    ui.setFrames(frames, FRAMES_SIZE);
    ui.setTimePerFrame(6000);

    ip = WiFi.localIP().toString();
    Serial.print("Camera Ready! Use 'http://");
    Serial.print(ip);
    Serial.println("' to connect");
}

void loop()
{
    if (ui.update()) {
        button1.tick();
        delay(500);
    }
}

